@@ -1,14 +1,14 @@
 from typing import List
 
 import pandas as pd  # noqa
-from truss.build import mk_truss
+from truss.build import create
 
 
-def test_infer_deps_through_mk_truss_local_import(sklearn_rfc_model, tmp_path):
+def test_infer_deps_through_create_local_import(sklearn_rfc_model, tmp_path):
     dir_path = tmp_path / "truss"
     import requests  # noqa
 
-    tr = mk_truss(
+    tr = create(
         sklearn_rfc_model,
         target_directory=dir_path,
     )
@@ -16,11 +16,9 @@ def test_infer_deps_through_mk_truss_local_import(sklearn_rfc_model, tmp_path):
     _validate_that_package_is_in_requirements(spec.requirements, "requests")
 
 
-def test_infer_deps_through_mk_truss_top_of_the_file_import(
-    sklearn_rfc_model, tmp_path
-):
+def test_infer_deps_through_create_top_of_the_file_import(sklearn_rfc_model, tmp_path):
     dir_path = tmp_path / "truss"
-    tr = mk_truss(
+    tr = create(
         sklearn_rfc_model,
         target_directory=dir_path,
     )
